"""Control timer to put speaker into sleep mode."""


class Timer:
    """
    Control timer to put speaker into sleep mode.
    """

    def __init__(self, api):
        """
        :param api: SamsungMultiroomApi instance
        """
        self._api = api

    def start(self, delay):
        """
        Set timer to put speaker into sleep mode after delay.

        :param delay: Delay in seconds
        """
        self._api.set_sleep_timer('start', delay)

    def stop(self):
        """
        Stop the timer.
        """
        self._api.set_sleep_timer('off', 0)

    def get_remaining_time(self):
        """
        :returns: Remaining timer seconds, or 0 if not enabled.
        """
        return int(self._api.get_sleep_timer()['sleeptime'])



"""Generic music streaming app service browser."""
from ...api import paginator
from ..browser import Browser
from ..browser import Item
from ..browser import path_to_folders


class AppBrowser(Browser):
    """
    Generic music streaming app service browser.
    """

    def __init__(self, api, app_id, app_name, path=None, items=None):
        super().__init__(path, items)

        self._api = api
        self._id = app_id
        self._name = app_name

    def get_name(self):
        return self._name

    def browse(self, path=None):
        folders = path_to_folders(path)

        items = self._get_initial_items(folders)
        depth = self._get_initial_depth(folders)

        for folder in folders:
            depth += 1

            # locate prepopulated items matching folder
            parent_id = next(iter([i.object_id for i in items if i.name == folder]), None)

            if parent_id is None:
                data_list = self._api.get_cp_submenu()
            elif depth <= 2:
                data_list = paginator(self._api.set_select_cp_submenu, parent_id, 0, 30)
            else:
                data_list = paginator(self._api.get_select_radio_list, self._api.get_current_radio_list, parent_id, 0,
                                      30)

            items = [self._factory_item(data) for data in data_list]

        return AppBrowser(self._api, self._id, self._name, self._merge_path(path), items)

    def _factory_item(self, data):
        kwargs = {'metadata': {}}

        # mandatory
        if '@id' in data:
            kwargs['object_id'] = data['@id']
        if 'contentid' in data:
            kwargs['object_id'] = data['contentid']

        if '@type' in data and data['@type'] in ('1', '2'):
            kwargs['object_type'] = 'app_audio'
        else:
            kwargs['object_type'] = 'container'

        if 'submenuitem_localized' in data:
            kwargs['name'] = data['submenuitem_localized']
        if 'title' in data:
            kwargs['name'] = data['title']

        # metadata
        if 'artist' in data:
            kwargs['metadata']['artist'] = data['artist']
        if 'album' in data:
            kwargs['metadata']['album'] = data['album']
        if 'tracklength' in data:
            kwargs['metadata']['duration'] = int(data['tracklength'])
        if 'thumbnail' in data:
            kwargs['metadata']['thumbnail_url'] = data['thumbnail']

        return Item(**kwargs)



"""DLNA service browser."""
from ...api import paginator
from ..browser import Browser
from ..browser import Item
from ..browser import path_to_folders


class DlnaBrowser(Browser):
    """
    DLNA DMS device browser.
    """

    def __init__(self, api, path=None, items=None):
        super().__init__(path, items)

        self._api = api

    def get_name(self):
        return 'dlna'

    def browse(self, path=None):
        folders = path_to_folders(path)

        items = self._get_initial_items(folders)

        for folder in folders:
            # locate prepopulated items matching folder
            device_udn, parent_id = next(
                iter([(i.device_udn, i.object_id) for i in items if i.name == folder]), (None, None))

            # if we don't have device udn we search through devices
            if device_udn is None:
                data_list = paginator(self._api.get_dms_list, 0, 20)
            elif parent_id is None:
                data_list = paginator(self._api.pc_get_music_list_by_category, device_udn, 0, 20)
            else:
                data_list = paginator(self._api.pc_get_music_list_by_id, device_udn, parent_id, 0, 20)

            items = [self._factory_item(data) for data in data_list]

        return DlnaBrowser(self._api, self._merge_path(path), items)

    def _factory_item(self, data):
        kwargs = {'metadata': {}}

        # mandatory
        if 'dmsid' in data:
            kwargs['object_id'] = None
        if '@object_id' in data:
            kwargs['object_id'] = data['@object_id']

        if 'type' in data and data['type'] == 'AUDIO':
            kwargs['object_type'] = 'dlna_audio'
        else:
            kwargs['object_type'] = 'container'

        if 'dmsname' in data:
            kwargs['name'] = data['dmsname']
        if 'title' in data:
            kwargs['name'] = data['title']

        # metadata
        if 'artist' in data:
            kwargs['metadata']['artist'] = data['artist']
        if 'album' in data:
            kwargs['metadata']['album'] = data['album']
        if 'timelength' in data and data['timelength']:
            (hours, minutes, seconds) = data['timelength'].split(':')
            kwargs['metadata']['duration'] = int(hours) * 3600 + int(minutes) * 60 + int(float(seconds))
        if 'thumbnail' in data:
            kwargs['metadata']['thumbnail_url'] = data['thumbnail']
        if 'thumbnail_JPG_LRG' in data:
            kwargs['metadata']['thumbnail_url'] = data['thumbnail_JPG_LRG']
        if 'devicetype' in data:
            kwargs['metadata']['device_type'] = data['devicetype']
        if 'dmsid' in data:
            kwargs['metadata']['device_udn'] = data['dmsid']
        if 'device_udn' in data:
            kwargs['metadata']['device_udn'] = data['device_udn']

        return Item(**kwargs)



import unittest
from unittest.mock import MagicMock
from unittest.mock import call

from samsung_multiroom.service import Item
from samsung_multiroom.service.app import AppBrowser


def _get_browser():
    api = MagicMock()
    api.get_cp_submenu.return_value = [
        {
            '@id': '0',
            'submenuitem_localized': 'Flow',
        },
        {
            '@id': '1',
            'submenuitem_localized': 'Browse',
        },
        {
            '@id': '2',
            'submenuitem_localized': 'Playlist picks',
        },
    ]
    api.set_select_cp_submenu.return_value = [
        {
            '@type': '0',
            'title': 'All',
            'contentid': '0',
        },
        {
            '@type': '0',
            'title': 'Pop',
            'contentid': '1',
        },
        {
            '@type': '0',
            'title': 'Rap/Hip Hop',
            'contentid': '2',
        },
        {
            '@type': '0',
            'title': 'Rock',
            'contentid': '3',
        },
        {
            '@type': '0',
            'title': 'Dance',
            'contentid': '4',
        },
    ]
    api.get_select_radio_list.side_effect = [
        [
            {
                '@type': '0',
                'title': 'Mixes',
                'contentid': '0',
            },
            {
                '@type': '0',
                'title': 'Playlists',
                'contentid': '1',
            },
            {
                '@type': '0',
                'title': 'Albums',
                'contentid': '2',
            },
            {
                '@type': '0',
                'title': 'Artists',
                'contentid': '3',
            },
            {
                '@type': '0',
                'title': 'Editor\'s Picks',
                'contentid': '4',
            },
        ],
        [
            {
                '@type': '3',
                'title': 'Queen',
                'contentid': '0',
                'thumbnail': 'http://api.deezer.com/artist/412/image',
            },
            {
                '@type': '3',
                'title': 'The Beatles',
                'contentid': '1',
                'thumbnail': 'http://api.deezer.com/artist/1/image',
            },
            {
                '@type': '3',
                'title': 'Linking Park',
                'contentid': '2',
                'thumbnail': 'http://api.deezer.com/artist/92/image',
            },
        ],
        [
            {
                '@type': '2',
                'artist': 'Queen',
                'album': None,
                'mediaid': '412',
                'title': 'Queen Mixes',
                'contentid': '0',
                'thumbnail': 'http://api.deezer.com/artist/412/image',
            },
            {
                '@cat': 'Top Tracks',
                '@type': '1',
                '@available': '1',
                '@currentplaying': '1',
                'artist': 'Queen',
                'album': 'A Night At The Opera (2011 Remaster)',
                'mediaid': '9997018',
                'tracklength': '358',
                'title': 'Queen Mixes',
                'contentid': '1',
                'thumbnail': 'http://api.deezer.com/album/915785/image',
            },
            {
                '@cat': 'Top Tracks',
                '@type': '1',
                '@available': '1',
                'artist': 'Queen',
                'album': 'Jazz (2011 Remaster)',
                'mediaid': '12209331',
                'tracklength': '209',
                'title': 'Don\'t Stop Me Now',
                'contentid': '2',
                'thumbnail': 'http://api.deezer.com/album/1121401/image',
            },
            {
                '@cat': 'Albums',
                '@type': '4',
                'artist': None,
                'title': 'Queen Forever',
                'contentid': '6',
                'thumbnail': 'http://api.deezer.com/album/8980335/image',
            },
        ]
    ]

    browser = AppBrowser(api, 3, 'service 3')

    return (browser, api)


class TestAppBrowser(unittest.TestCase):

    def test_get_name(self):
        browser, api = _get_browser()

        name = browser.get_name()

        self.assertEqual(name, 'service 3')

    def test_browse_from_root(self):
        browser, api = _get_browser()

        browser = browser.browse()

        api.get_cp_submenu.assert_called_once()

        self.assertEqual(browser.get_path(), '/')
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'container')
        self.assertEqual(browser[0].name, 'Flow')

    @unittest.mock.patch('inspect.signature')
    def test_browse_second_level(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
        ]

        browser, api = _get_browser()

        browser = browser.browse('/Browse')

        api.get_cp_submenu.assert_called_once()
        api.set_select_cp_submenu.assert_called_once_with(content_id='1', start_index=0, list_count=30)

        self.assertEqual(browser.get_path(), '/Browse')
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'container')
        self.assertEqual(browser[0].name, 'All')

    @unittest.mock.patch('inspect.signature')
    def test_browse_full_level(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
        ]

        browser, api = _get_browser()

        browser = browser.browse('/Browse/Rock/Artists/Queen/')

        api.get_cp_submenu.assert_called_once()
        api.set_select_cp_submenu.assert_called_once_with(content_id='1', start_index=0, list_count=30)
        api.get_select_radio_list.assert_has_calls([
            call(content_id='3', start_index=0, list_count=30),
            call(content_id='3', start_index=0, list_count=30),
            call(content_id='0', start_index=0, list_count=30),
        ])

        self.assertEqual(browser.get_path(), '/Browse/Rock/Artists/Queen')
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'app_audio')
        self.assertEqual(browser[0].name, 'Queen Mixes')

    @unittest.mock.patch('inspect.signature')
    def test_browse_relative_path(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
        ]

        browser, api = _get_browser()

        browser = browser.browse('/Browse/Rock').browse('Artists/Queen/')

        api.get_cp_submenu.assert_called_once()
        api.set_select_cp_submenu.assert_called_once_with(content_id='1', start_index=0, list_count=30)
        api.get_select_radio_list.assert_has_calls([
            call(content_id='3', start_index=0, list_count=30),
            call(content_id='3', start_index=0, list_count=30),
            call(content_id='0', start_index=0, list_count=30),
        ])

        self.assertEqual(browser.get_path(), '/Browse/Rock/Artists/Queen')
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'app_audio')
        self.assertEqual(browser[0].name, 'Queen Mixes')



"""Generic music streaming app service player."""
from ..player import REPEAT_OFF
from ..player import Player
from ..player import Track
from ..player import init_track_kwargs
from ..player import unsupported


class AppPlayer(Player):
    """Controls player in WIFI+cp mode."""

    def __init__(self, api):
        self._api = api

    def play(self, playlist):
        """
        Enqueue and play a playlist.

        Playlist items must be an object with following attributes:
        - object_id - object id
        - object_type - must be 'app_audio'

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        item_ids = []
        for track in playlist:
            if track.object_type not in ['app_audio']:
                continue

            item_ids.append(track.object_id)

        if item_ids:
            self._api.set_play_select(item_ids)
            return True

        return False

    @unsupported
    def jump(self, time):
        """
        Not supported for app.

        :param time: Time from the beginning of the track in seconds
        """

    def resume(self):
        """Play/resume current track."""
        self._api.set_playback_control('play')

    @unsupported
    def stop(self):
        """Stop current track and reset position to the beginning."""
        raise NotImplementedError()

    def pause(self):
        """Pause current track and retain position."""
        self._api.set_playback_control('pause')

    def next(self):
        """Play next track in the queue."""
        self._api.set_skip_current_track()

    def previous(self):
        """Play previous track in the queue."""
        playlist = self._api.get_cp_player_playlist()

        for i, playlist_item in enumerate(playlist):
            if '@currentplaying' in playlist_item and playlist_item['@currentplaying'] == '1':
                previous_index = max(0, i - 1)

                self._api.set_play_cp_playlist_track(playlist[previous_index]['contentid'])

    @unsupported
    def repeat(self, mode):
        """Not supported for apps."""

    @unsupported
    def shuffle(self, enabled):
        """Not supported for apps."""

    def get_repeat(self):
        """Not supported for apps."""
        return REPEAT_OFF

    def get_shuffle(self):
        """Not supported for apps."""
        return False

    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        playlist = self._api.get_cp_player_playlist()
        try:
            playlist_item = [p for p in playlist if '@currentplaying' in p and p['@currentplaying'] == '1'][0]
        except KeyError:
            return None

        track_kwargs = init_track_kwargs('app_audio')

        if 'title' in playlist_item:
            track_kwargs['title'] = playlist_item['title']
        if 'artist' in playlist_item:
            track_kwargs['artist'] = playlist_item['artist']
        if 'album' in playlist_item:
            track_kwargs['album'] = playlist_item['album']
        if 'thumbnail' in playlist_item and 'http' in playlist_item['thumbnail']:
            track_kwargs['thumbnail_url'] = playlist_item['thumbnail']
        if 'mediaid' in playlist_item:
            track_kwargs['metadata']['object_id'] = playlist_item['mediaid']

        play_time = self._api.get_current_play_time()

        if 'timelength' in play_time and play_time['timelength'] is not None:
            track_kwargs['duration'] = int(play_time['timelength'])
        if 'playtime' in play_time:
            track_kwargs['position'] = int(play_time['playtime'])

        return Track(**track_kwargs)

    def is_active(self, function, submode=None):
        """
        Check if this player is active based on current function/submode.

        :returns: Boolean True if function/submode is supported
        """
        return function == 'wifi' and submode == 'cp'



"""Event."""
from ...service.player import REPEAT_ALL
from ...service.player import REPEAT_OFF
from ...service.player import REPEAT_ONE
from ..event import Event


class SpeakerPlayerRepeatChangedEvent(Event):
    """Event when player's repeat state changes."""

    def __init__(self, repeat):
        """
        :param repeat: one of REPEAT_ constants
        """
        super().__init__('speaker.player.repeat.changed')

        self._repeat = repeat

    @property
    def repeat(self):
        """
        :returns: repeat state, True if repeat is on
        """
        return self._repeat

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerPlayerRepeatChangedEvent instance or None if response is unsupported
        """
        if response.name != 'RepeatMode':
            return None

        mode_map = {
            'all': REPEAT_ALL,
            'one': REPEAT_ONE,
            'off': REPEAT_OFF,
        }

        return cls(mode_map[response.data['repeat']])



"""Generic event."""


class Event:
    """Generic event."""

    def __init__(self, name):
        """
        :param name: Name of event as defined in Events class
        """
        self._name = name

    @property
    def name(self):
        """
        :returns: Name of event
        """
        return self._name



"""Event."""
from ..event import Event


class SpeakerMuteChangedEvent(Event):
    """Event when speaker mute state changes."""

    def __init__(self, muted):
        """
        :param muted: Boolean
        """
        super().__init__('speaker.mute.changed')

        self._muted = muted

    @property
    def muted(self):
        """
        :returns: Mute state, True if muted
        """
        return self._muted

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerMuteChangedEvent instance or None if response is unsupported
        """
        if response.name != 'MuteStatus':
            return None

        return cls(response.data['mute'] == 'on')



import unittest

from samsung_multiroom.service import Player
from samsung_multiroom.service.player import unsupported


class FakePlayer(Player):

    def play(self, playlist):
        """Fake."""
        return False

    def jump(self, time):
        """Fake."""

    def resume(self):
        """Fake."""

    def stop(self):
        """Fake."""

    def pause(self):
        """Fake."""

    def next(self):
        """Fake."""

    def previous(self):
        """Fake."""

    @unsupported
    def repeat(self, mode):
        """Fake."""

    def shuffle(self, enabled):
        """Fake."""

    def get_repeat(self):
        """Fake."""
        return REPEAT_OFF

    def get_shuffle(self):
        """Fake."""
        return False

    def get_current_track(self):
        """Fake."""
        return None

    def is_active(self, function, submode=None):
        """Fake."""
        return True



class TestPlayer(unittest.TestCase):

    def test_unsupported(self):
        player = FakePlayer()
        self.assertTrue(player.is_play_supported())
        self.assertFalse(player.is_repeat_supported())

        self.assertTrue(hasattr(player, 'is_play_supported'))
        self.assertTrue(hasattr(player, 'is_repeat_supported'))
        self.assertFalse(hasattr(player, 'is_nonsense_supported'))



import unittest

from samsung_multiroom.service import path_to_folders


class TestPathToFolders(unittest.TestCase):

    def test_path_to_folders(self):
        self.assertEqual(path_to_folders(None), [None])
        self.assertEqual(path_to_folders(''), [None])
        self.assertEqual(path_to_folders('/'), [None])
        self.assertEqual(path_to_folders('/Folder1'), [None, 'Folder1'])
        self.assertEqual(path_to_folders('/Folder1/'), [None, 'Folder1'])
        self.assertEqual(path_to_folders('/Folder1/Folder2/Folder3/'), [None, 'Folder1', 'Folder2', 'Folder3'])



import unittest
from unittest.mock import MagicMock
from unittest.mock import call

from samsung_multiroom.service import Item
from samsung_multiroom.service.dlna import DlnaBrowser


def get_dms_list_return_value():
    return [
        {
            '@device_id': '0',
            'dmsid': 'uuid:00113249-398f-0011-8f39-8f3949321100',
            'dmsname': 'NAS',
            'devicetype': 'network',
            'thumbnail_PNG_LRG': 'http://192.168.1.111:50001/tmp_icon/dmsicon120.png',
            'thumbnail_JPG_LRG': 'http://192.168.1.111:50001/tmp_icon/dmsicon120.jpg',
            'thumbnail_PNG_SM': 'http://192.168.1.111:50001/tmp_icon/dmsicon48.png',
            'thumbnail_JPG_SM': 'http://192.168.1.111:50001/tmp_icon/dmsicon48.jpg',
        },
    ]

def pc_get_music_list_by_category_return_value():
    return [
        {
            '@object_id': '21',
            'type': 'CONTAINER',
            'playindex': '-1',
            'name': None,
            'title': 'Music',
            'artist': None,
            'album': None,
            'thumbnail': None,
            'timelength': None,
            'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
        },
        {
            '@object_id': '37',
            'type': 'CONTAINER',
            'playindex': '-1',
            'name': None,
            'title': 'Photo',
            'artist': None,
            'album': None,
            'thumbnail': None,
            'timelength': None,
            'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
        },
        {
            '@object_id': '44',
            'type': 'CONTAINER',
            'playindex': '-1',
            'name': None,
            'title': 'Video',
            'artist': None,
            'album': None,
            'thumbnail': None,
            'timelength': None,
            'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
        },
    ]

def pc_get_music_list_by_id_side_effect(device_udn, parent_id, start_index, list_count):
    if parent_id == '21':
        return [
            {
                '@object_id': '22',
                'type': 'CONTAINER',
                'playindex': '-1',
                'name': None,
                'title': 'By Folder',
                'artist': None,
                'album': None,
                'thumbnail': None,
                'timelength': None,
                'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
            },
        ]
    if parent_id == '22':
        return [
            {
                '@object_id': '22$@52941',
                'type': 'AUDIO',
                'playindex': '0',
                'name': 'La femme d\'argent.mp3',
                'title': 'La femme d\'argent',
                'artist': 'Air',
                'album': 'Moon Safari',
                'thumbnail': 'http://192.168.1.111:50002/transcoder/jpegtnscaler.cgi/folderart/52941.jpg',
                'timelength': '0:07:11.000',
                'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
            },
            {
                '@object_id': '22$@52942',
                'type': 'AUDIO',
                'playindex': '0',
                'name': 'Sexy boy.mp3',
                'title': 'Sexy boy',
                'artist': 'Air',
                'album': 'Moon Safari',
                'thumbnail': 'http://192.168.1.111:50002/transcoder/jpegtnscaler.cgi/folderart/52942.jpg',
                'timelength': '0:04:58.000',
                'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
            },
        ]


class TestDlnaBrowser(unittest.TestCase):

    @unittest.mock.patch('inspect.signature')
    def test_browse_from_root(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.get_dms_list.return_value = get_dms_list_return_value()

        browser = DlnaBrowser(api)
        browser = browser.browse()

        api.get_dms_list.assert_called_once_with(start_index=0, list_count=20)

        self.assertEqual(browser.get_path(), '/')
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, None)
        self.assertEqual(browser[0].object_type, 'container')
        self.assertEqual(browser[0].name, 'NAS')
        self.assertEqual(browser[0].device_udn, 'uuid:00113249-398f-0011-8f39-8f3949321100')

    @unittest.mock.patch('inspect.signature')
    def test_browse_second_level(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.get_dms_list.return_value = get_dms_list_return_value()
        api.pc_get_music_list_by_category.return_value = pc_get_music_list_by_category_return_value()

        browser = DlnaBrowser(api)
        browser = browser.browse('/NAS/')

        api.get_dms_list.assert_called_once_with(start_index=0, list_count=20)
        api.pc_get_music_list_by_category.assert_called_once_with(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', start_index=0, list_count=20)

        self.assertEqual(browser.get_path(), '/NAS')
        self.assertEqual(len(browser), 3)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '21')
        self.assertEqual(browser[0].object_type, 'container')
        self.assertEqual(browser[0].name, 'Music')
        self.assertEqual(browser[0].device_udn, 'uuid:00113249-398f-0011-8f39-8f3949321100')

    @unittest.mock.patch('inspect.signature')
    def test_browse_full_level(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.get_dms_list.return_value = get_dms_list_return_value()
        api.pc_get_music_list_by_category.return_value = pc_get_music_list_by_category_return_value()
        api.pc_get_music_list_by_id.side_effect = pc_get_music_list_by_id_side_effect

        browser = DlnaBrowser(api)
        browser = browser.browse('/NAS/Music/By Folder')

        api.get_dms_list.assert_called_once_with(start_index=0, list_count=20)
        api.pc_get_music_list_by_category.assert_called_once_with(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', start_index=0, list_count=20)
        api.pc_get_music_list_by_id.assert_has_calls([
            call(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', parent_id='21', start_index=0, list_count=20),
            call(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', parent_id='22', start_index=0, list_count=20),
        ])

        self.assertEqual(browser.get_path(), '/NAS/Music/By Folder')
        self.assertEqual(len(browser), 2)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '22$@52941')
        self.assertEqual(browser[0].object_type, 'dlna_audio')
        self.assertEqual(browser[0].name, 'La femme d\'argent')
        self.assertEqual(browser[0].device_udn, 'uuid:00113249-398f-0011-8f39-8f3949321100')

    @unittest.mock.patch('inspect.signature')
    def test_browse_relative_path(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'device_udn': None, 'parent_id':None, 'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.get_dms_list.return_value = get_dms_list_return_value()
        api.pc_get_music_list_by_category.return_value = pc_get_music_list_by_category_return_value()
        api.pc_get_music_list_by_id.side_effect = pc_get_music_list_by_id_side_effect

        browser = DlnaBrowser(api)
        browser = browser.browse('/NAS/Music/').browse('By Folder')

        api.get_dms_list.assert_called_once_with(start_index=0, list_count=20)
        api.pc_get_music_list_by_category.assert_called_once_with(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', start_index=0, list_count=20)
        api.pc_get_music_list_by_id.assert_has_calls([
            call(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', parent_id='21', start_index=0, list_count=20),
            call(device_udn='uuid:00113249-398f-0011-8f39-8f3949321100', parent_id='22', start_index=0, list_count=20),
        ])

        self.assertEqual(browser.get_path(), '/NAS/Music/By Folder')
        self.assertEqual(len(browser), 2)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].object_id, '22$@52941')
        self.assertEqual(browser[0].object_type, 'dlna_audio')
        self.assertEqual(browser[0].name, 'La femme d\'argent')
        self.assertEqual(browser[0].device_udn, 'uuid:00113249-398f-0011-8f39-8f3949321100')



"""Event dispatching."""
import fnmatch


class EventLoop:
    """
    Listen to speaker events.

    Use add_listener to subscribe to events of particular type.
    """

    def __init__(self, api_stream):
        """
        :param api_stream: ApiStream instance
        """
        self._api_stream = api_stream
        self._listeners = []
        self._factories = _get_default_factories()

    def register_factory(self, factory):
        """
        Register event factory function.

        :param factory: Callable accepting ApiResponse instance
        """
        if not callable(factory):
            raise ValueError('factory must be callable')

        self._factories.append(factory)

    def add_listener(self, event_name, listener):
        """
        :param event_name: Event name. See Events for all supported events
        :param listener: Callable. Will be called on a matching event and passed matching Event object
        """
        if not callable(listener):
            raise ValueError('listener must be a callable')

        self._listeners.append((event_name, listener))

    async def loop(self):
        """
        Start emitting speaker events.
        """
        for response in self._api_stream.open('/UIC?cmd=%3Cname%3EGetMainInfo%3C/name%3E'):
            event = self._factory(response)
            if event:
                self._dispatch_event(event)

    def _dispatch_event(self, event):
        for event_name, listener in self._listeners:
            if fnmatch.fnmatch(event.name, event_name):
                listener(event)

    def _factory(self, response):
        """
        Factory event from response.

        :param response: ApiResponse instance
        """
        for factory in reversed(self._factories):
            event = factory(response)

            if event:
                return event

        return None


def _get_default_factories():
    event_classes = [
        ('samsung_multiroom.event.type.speaker_mute_changed', 'SpeakerMuteChangedEvent'),
        ('samsung_multiroom.event.type.speaker_player', 'SpeakerPlayerEvent'),
        ('samsung_multiroom.event.type.speaker_player_repeat_changed', 'SpeakerPlayerRepeatChangedEvent'),
        ('samsung_multiroom.event.type.speaker_player_shuffle_changed', 'SpeakerPlayerShuffleChangedEvent'),
        ('samsung_multiroom.event.type.speaker_service', 'SpeakerServiceEvent'),
        ('samsung_multiroom.event.type.speaker_volume_changed', 'SpeakerVolumeChangedEvent'),
    ]

    factories = []

    for module_name, class_name in event_classes:
        mod = __import__(module_name, fromlist=[class_name])
        klass = getattr(mod, class_name)

        factories.append(klass.factory)

    return factories



"""Select the right player for the current source."""
from .player import REPEAT_OFF
from .player import Player
from .player import get_is_supported_function_name
from .player import unsupported


class PlayerOperator(Player):
    """Select the right player for the current source."""

    def __init__(self, api, players=None):
        """
        Initialise player operator.

        :param api: SamsungMultiroomApi instance
        :param players: List of Player instances
        """
        self._api = api
        self._players = []

        for player in (players or []):
            self.add_player(player)

    def add_player(self, player):
        """
        Add player.

        :param player: Player instance
        """
        if not isinstance(player, Player):
            raise ValueError('Player must be a subclass of Player class')

        self._players.append(player)

    def get_player(self):
        """
        Get currently active player based on selected source.

        :returns: Player instance
        """
        func = self._api.get_func()

        function = func['function']
        submode = func['submode']

        for player in self._players:
            if player.is_active(function, submode):
                return player

        return NullPlayer()

    def play(self, playlist):
        """
        Find a suitable player and play a playlist.

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        for player in self._players:
            if player.play(playlist):
                return True

        return False

    def jump(self, time):
        """
        Advance current playback to specific time.

        :param time: Time from the beginning of the track in seconds
        """
        self.get_player().jump(time)

    def resume(self):
        """Play/resume current track."""
        self.get_player().resume()

    def stop(self):
        """Stop current track and reset position to the beginning."""
        self.get_player().stop()

    def pause(self):
        """Pause current track and retain position."""
        self.get_player().pause()

    def next(self):
        """Play next track in the queue."""
        self.get_player().next()

    def previous(self):
        """Play previous track in the queue."""
        self.get_player().previous()

    def repeat(self, mode):
        """
        Set playback repeat mode.

        :param mode: one of REPEAT_* constants
        """
        self.get_player().repeat(mode)

    def shuffle(self, enabled):
        """
        Enable/disable playback shuffle mode.

        :param enabled: True to enable, False to disable
        """
        self.get_player().shuffle(enabled)

    def get_repeat(self):
        """
        Get playback repeat mode.

        :returns: one of REPEAT_* constants
        """
        return self.get_player().get_repeat()

    def get_shuffle(self):
        """
        Get playback shuffle mode.

        :returns: boolean, True if enabled, False otherwise
        """
        return self.get_player().get_shuffle()

    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        return self.get_player().get_current_track()

    def is_active(self, function, submode=None):
        """
        Check if this player is active based on current function/submode.

        :returns: Boolean True if function/submode is supported
        """
        for player in self._players:
            if player.is_active(function, submode):
                return True

        return False

    def __getattribute__(self, name):
        """
        Delegates is_[function_name]_supported to the currently active player.
        """
        if get_is_supported_function_name(name) is None:
            return super().__getattribute__(name)

        return getattr(self.get_player(), name)


class NullPlayer(Player):
    """Catch all player if no others supported current function."""

    @unsupported
    def play(self, playlist):
        """
        Null player is unable to play anything.

        :param playlist: Playlist instance
        :returns: Boolean False
        """
        return False

    @unsupported
    def jump(self, time):
        """Do nothing."""

    @unsupported
    def resume(self):
        """Do nothing."""

    @unsupported
    def stop(self):
        """Do nothing."""

    @unsupported
    def pause(self):
        """Do nothing."""

    @unsupported
    def next(self):
        """Do nothing."""

    @unsupported
    def previous(self):
        """Do nothing."""

    @unsupported
    def repeat(self, mode):
        """Do nothing."""

    @unsupported
    def shuffle(self, enabled):
        """Do nothing."""

    def get_repeat(self):
        """Always off."""
        return REPEAT_OFF

    def get_shuffle(self):
        """Always False."""
        return False

    def get_current_track(self):
        """
        Do nothing.

        :returns: None
        """
        return None

    def is_active(self, function, submode=None):
        """
        Check if this player is active based on current function/submode.

        :returns: Boolean True
        """
        return True



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.type.speaker_player import SpeakerPlayerEvent


class TestSpeakerPlayerEvent(unittest.TestCase):

    def test_factory(self):
        responses = {
            'StartPlaybackEvent': 'speaker.player.playback_started',
            'StopPlaybackEvent': 'speaker.player.playback_ended',
            'EndPlaybackEvent': 'speaker.player.playback_ended',
            'PausePlaybackEvent': 'speaker.player.playback_paused',
            'MediaBufferStartEvent': 'speaker.player.buffering_started',
            'MediaBufferEndEvent': 'speaker.player.buffering_ended',
            'OtherResponse': None,
        }

        for response_name, event_name in responses.items():
            response = MagicMock()
            response.name = response_name

            event = SpeakerPlayerEvent.factory(response)

            if event_name:
                self.assertIsInstance(event, SpeakerPlayerEvent)
                self.assertEqual(event.name, event_name)
            else:
                self.assertIsNone(event)



"""Player allows playback control depending on selected source."""
import abc

# repeat mode constants
REPEAT_ONE = 'one'
REPEAT_ALL = 'all'
REPEAT_OFF = 'off'


class Player(metaclass=abc.ABCMeta):
    """Player interface to control playback functions."""

    @abc.abstractmethod
    def play(self, playlist):
        """
        Enqueue and play a playlist.

        Player may choose to not play the playlist if it's not compatible with this player. For instance you can't
        play DLNA source tracks using TuneIn player. If player is unable to play the playlist it must return False.

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def jump(self, time):
        """
        Advance current playback to specific time.

        :param time: Time from the beginning of the track in seconds
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def resume(self):
        """Play/resume current track."""
        raise NotImplementedError()

    @abc.abstractmethod
    def stop(self):
        """Stop current track and reset position to the beginning."""
        raise NotImplementedError()

    @abc.abstractmethod
    def pause(self):
        """Pause current track and retain position."""
        raise NotImplementedError()

    @abc.abstractmethod
    def next(self):
        """Play next track in the queue."""
        raise NotImplementedError()

    @abc.abstractmethod
    def previous(self):
        """Play previous track in the queue."""
        raise NotImplementedError()

    @abc.abstractmethod
    def repeat(self, mode):
        """
        Set playback repeat mode.

        :param mode: one of REPEAT_* constants
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def shuffle(self, enabled):
        """
        Enable/disable playback shuffle mode.

        :param enabled: True to enable, False to disable
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_repeat(self):
        """
        Get playback repeat mode.

        :returns: one of REPEAT_* constants
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_shuffle(self):
        """
        Get playback shuffle mode.

        :returns: boolean, True if enabled, False otherwise
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def is_active(self, function, submode=None):
        """
        Check if this player is active based on current function/submode.

        :returns: Boolean True if function/submode is supported
        """
        raise NotImplementedError()

    def __getattribute__(self, name):
        """
        Magic is_[function]_supported method.

        Function can be any Player method. In order to mark method as unsupported, use @unsupported decorator.

        Example:
            MyPlayer(Player):
                @unsupported
                def play(self, playlist):
                    return False

            player = MyPlayer()
            player.is_play_supported() # returns False
        """
        try:
            return super().__getattribute__(name)
        except AttributeError:
            function_name = get_is_supported_function_name(name)

            if not function_name:
                raise

            if not hasattr(self, function_name):
                raise

            function = getattr(self, function_name)
            if not hasattr(function, '__is_supported__'):
                return lambda: True

            return lambda: bool(function.__is_supported__)


class Track:
    """Defines a media track on the playlist."""

    def __init__(self, title, artist, album, duration, position, thumbnail_url, metadata=None):
        self._title = title
        self._artist = artist
        self._album = album
        self._duration = duration
        self._position = position
        self._thumbnail_url = thumbnail_url
        self._metadata = metadata or {}

    @property
    def title(self):
        """
        :returns: Title of the current track
        """
        return self._title

    @property
    def artist(self):
        """
        :returns: Artist of the current track
        """
        return self._artist

    @property
    def album(self):
        """
        :returns: Album title of the current track
        """
        return self._album

    @property
    def duration(self):
        """
        :returns: Duration in seconds
        """
        return self._duration

    @property
    def position(self):
        """
        :returns: Current playback position in seconds
        """
        return self._position

    @property
    def thumbnail_url(self):
        """
        :returns: URL of the track thumbnail
        """
        return self._thumbnail_url

    def __getattr__(self, name):
        """
        :returns: Metadata item value
        """
        if name in self._metadata:
            return self._metadata[name]

        return None


def init_track_kwargs(object_type):
    """
    :returns: kwargs dict fro Track initialisation
    """
    return {
        'title': None,
        'artist': None,
        'album': None,
        'duration': None,
        'position': None,
        'thumbnail_url': None,
        'metadata': {
            'object_id': None,
            'object_type': object_type,
        }
    }


def unsupported(function):
    """Decorator to mark player function as unsupported."""
    function.__is_supported__ = False

    return function


def get_is_supported_function_name(name):
    """
    :param name: function name
    :returns: Function name from is_[function_name]_supported structure, None otherwise
    """
    import re

    pattern = re.compile(r'^is_(\w+)_supported$')
    matches = pattern.findall(name)

    if not matches:
        return None

    return matches[0]



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import REPEAT_ALL
from samsung_multiroom.service import REPEAT_OFF
from samsung_multiroom.service.tunein import TuneInPlayer


def _get_player():
    api = MagicMock()
    api.get_preset_list.return_value = [
        {
            'kind': 'speaker',
            'title': 'Radio 1',
            'description': 'Radio 1 description',
            'thumbnail': 'http://radio1.org/thumbnail.png',
            'contentid': '0',
            'mediaid': '1111',
        },
        {
            'kind': 'speaker',
            'title': 'Radio 2',
            'description': 'Radio 2 description',
            'thumbnail': 'http://radio2.org/thumbnail.png',
            'contentid': '1',
            'mediaid': '2222',
        },
        {
            'kind': 'speaker',
            'title': 'Radio 3',
            'description': 'Radio 3 description',
            'thumbnail': 'http://radio3.org/thumbnail.png',
            'contentid': '2',
            'mediaid': '3333',
        },
        {
            'kind': 'my',
            'title': 'Radio 4',
            'description': 'Radio 4 description',
            'thumbnail': 'http://radio4.org/thumbnail.png',
            'contentid': '3',
            'mediaid': '4444',
        },
        {
            'kind': 'my',
            'title': 'Radio 5',
            'description': 'Radio 5 description',
            'thumbnail': 'http://radio5.org/thumbnail.png',
            'contentid': '4',
            'mediaid': '5555',
        },
    ]
    api.get_radio_info.return_value = {
        'cpname': 'TuneIn',
        'root': 'Favorites',
        'presetindex': '0',
        'title': 'Radio 1',
        'description': 'Radio 1 description',
        'thumbnail': 'http://radio1.org/thumbnail.png',
        'mediaid': '1111',
        'allowfeedback': '0',
        'timestamp': '2018-12-28T18:07:07Z',
        'no_queue': '1',
        'playstatus': 'play',
    }

    player = TuneInPlayer(api)

    return (player, api)


class TestTuneInPlayer(unittest.TestCase):

    def test_is_supported(self):
        player, api = _get_player()

        self.assertTrue(player.is_play_supported())
        self.assertFalse(player.is_jump_supported())
        self.assertTrue(player.is_resume_supported())
        self.assertFalse(player.is_stop_supported())
        self.assertTrue(player.is_pause_supported())
        self.assertTrue(player.is_next_supported())
        self.assertTrue(player.is_previous_supported())
        self.assertFalse(player.is_repeat_supported())
        self.assertFalse(player.is_shuffle_supported())

    def test_play(self):
        playlist = [
            type('Item', (object, ), {
                'object_id': '1',
                'object_type': 'some_type',
                'title': 'title 1',
            }),
            type('Item', (object, ), {
                'object_id': '2',
                'object_type': 'tunein_radio',
                'title': 'radio 2',
            }),
            type('Item', (object, ), {
                'object_id': '3',
                'object_type': 'tunein_radio',
                'title': 'radio 3',
            }),
            type('Item', (object, ), {
                'object_id': '4',
                'object_type': 'some_type2',
                'title': 'title 4',
            })
        ]

        player, api = _get_player()

        player.play(playlist)

        api.set_play_select.assert_called_once_with('2')

    def test_play_returns_false_for_unsupported_playlist(self):
        playlist = [
            type('Item', (object, ), {
                'object_id': '1',
                'object_type': 'some_type',
                'title': 'title 1',
            }),
            type('Item', (object, ), {
                'object_id': '4',
                'object_type': 'some_type2',
                'title': 'title 4',
            })
        ]

        player, api = _get_player()

        self.assertFalse(player.play(playlist))

        api.set_play_select.assert_not_called()

    def test_jump(self):
        player, api = _get_player()

        player.jump(50)

        api.set_search_time.assert_not_called()

    def test_resume(self):
        player, api = _get_player()

        player.resume()

        api.set_select_radio.assert_called_once()

    @unittest.skip('Pending implementation')
    def test_stop(self):
        player, api = _get_player()

        player.stop()

    def test_pause(self):
        player, api = _get_player()

        player.pause()

        api.set_playback_control.assert_called_once_with('pause')

    @unittest.mock.patch('inspect.signature')
    def test_next(self, signature):
        signature.return_value = type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}})

        player, api = _get_player()

        player.next()

        api.get_preset_list.assert_called_once_with(start_index=0, list_count=30)
        api.get_radio_info.assert_called_once()
        api.set_play_preset.assert_called_once_with(1, 1)
        api.set_select_radio.assert_called_once()

    @unittest.mock.patch('inspect.signature')
    def test_previous(self, signature):
        signature.return_value = type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}})

        player, api = _get_player()

        player.previous()

        api.get_preset_list.assert_called_once_with(start_index=0, list_count=30)
        api.get_radio_info.assert_called_once()
        api.set_play_preset.assert_called_once_with(0, 4)
        api.set_select_radio.assert_called_once()

    def test_repeat(self):
        player, api = _get_player()

        player.repeat(REPEAT_ALL)

        api.set_repeat_mode.assert_not_called()

    def test_shuffle(self):
        player, api = _get_player()

        player.shuffle(True)

        api.set_shuffle_mode.assert_not_called()

    def test_get_repeat(self):
        player, api = _get_player()

        repeat = player.get_repeat()

        self.assertEqual(repeat, REPEAT_OFF)

        api.get_repeat_mode.assert_not_called()

    def test_get_shuffle(self):
        player, api = _get_player()

        shuffle = player.get_shuffle()

        self.assertFalse(shuffle)

        api.get_repeat_mode.assert_not_called()

    def test_get_current_track(self):
        player, api = _get_player()

        track = player.get_current_track()

        api.get_radio_info.assert_called_once()

        self.assertEqual(track.title, 'Radio 1 description')
        self.assertEqual(track.artist, 'Radio 1')
        self.assertEqual(track.album, None)
        self.assertEqual(track.duration, None)
        self.assertEqual(track.position, None)
        self.assertEqual(track.thumbnail_url, 'http://radio1.org/thumbnail.png')
        self.assertEqual(track.object_id, None)
        self.assertEqual(track.object_type, 'tunein_radio')

    def test_is_active(self):
        player, api = _get_player()

        self.assertTrue(player.is_active('wifi', 'cp'))
        self.assertFalse(player.is_active('wifi', 'dlna'))
        self.assertFalse(player.is_active('bt'))



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import REPEAT_ALL
from samsung_multiroom.service import REPEAT_OFF
from samsung_multiroom.service.app import AppPlayer


def _get_player():
    api = MagicMock()

    player = AppPlayer(api)

    return (player, api)


class TestAppPlayer(unittest.TestCase):

    def test_is_supported(self):
        player, api = _get_player()

        self.assertTrue(player.is_play_supported())
        self.assertFalse(player.is_jump_supported())
        self.assertTrue(player.is_resume_supported())
        self.assertFalse(player.is_stop_supported())
        self.assertTrue(player.is_pause_supported())
        self.assertTrue(player.is_next_supported())
        self.assertTrue(player.is_previous_supported())
        self.assertFalse(player.is_repeat_supported())
        self.assertFalse(player.is_shuffle_supported())

    def test_play(self):
        player, api = _get_player()

        playlist = [
            type('Item', (object, ), {
                'object_id': 'id1',
                'object_type': 'some_type',
                'title': 'title 1',
            }),
            type('Item', (object, ), {
                'device_udn': 'device_udn',
                'object_id': '2',
                'object_type': 'app_audio',
                'title': 'track 2',
                'artist': 'artist 2',
                'thumbnail_url': 'thumb 2'
            }),
            type('Item', (object, ), {
                'device_udn': 'device_udn',
                'object_id': '3',
                'object_type': 'app_audio',
                'title': 'track 3',
                'artist': 'artist 3',
                'thumbnail_url': 'thumb 3'
            }),
            type('Item', (object, ), {
                'object_id': 'id4',
                'object_type': 'some_type2',
                'title': 'title 4',
            })
        ]

        player.play(playlist)

        api.set_play_select.assert_called_once_with(['2', '3'])

    @unittest.skip('No access to app supporting this feature')
    def test_jump(self):
        pass

    def test_resume(self):
        player, api = _get_player()

        player.resume()

        api.set_playback_control.assert_called_once_with('play')

    @unittest.skip('No access to app supporting this feature')
    def test_stop(self):
        pass

    def test_pause(self):
        player, api = _get_player()

        player.pause()

        api.set_playback_control.assert_called_once_with('pause')

    def test_next(self):
        player, api = _get_player()

        player.next()

        api.set_skip_current_track.assert_called_once()

    def test_previous(self):
        player, api = _get_player()

        api.get_cp_player_playlist.return_value = [
            # we don't care about other attributes
            {
                'contentid': '0',
            },
            {
                'contentid': '1',
            },
            {
                '@currentplaying': '1',
                'contentid': '2',
            },
            {
                'contentid': '3',
            }
        ]

        player.previous()

        api.set_play_cp_playlist_track.assert_called_once_with('1')

    def test_repeat(self):
        player, api = _get_player()

        player.repeat(REPEAT_ALL)

        api.set_repeat_mode.assert_not_called()

    def test_shuffle(self):
        player, api = _get_player()

        player.shuffle(True)

        api.set_shuffle_mode.assert_not_called()

    def test_get_repeat(self):
        player, api = _get_player()

        repeat = player.get_repeat()

        self.assertEqual(repeat, REPEAT_OFF)

        api.get_repeat_mode.assert_not_called()

    def test_get_shuffle(self):
        player, api = _get_player()

        shuffle = player.get_shuffle()

        self.assertFalse(shuffle)

        api.get_repeat_mode.assert_not_called()

    def test_get_current_track(self):
        player, api = _get_player()

        api.get_cp_player_playlist.return_value = [
            {
                '@type': '1',
                '@available': '1',
                '@currentplaying': '1',
                'artist': 'Madeleine Peyroux',
                'album': 'Careless Love',
                'mediaid': '881851',
                'tracklength': '0',
                'title': 'Don\'t Wait Too Long',
                'contentid': '0',
                'thumbnail': 'http://api.deezer.com/album/100127/image',
            },
            {
                '@type': '1',
                '@available': '1',
                'artist': 'artist',
                'album': 'album',
                'mediaid': 'mediaid',
                'tracklength': '0',
                'title': 'title',
                'contentid': '1',
                'thumbnail': 'http://thumb.url/image',
            },
            {
                '@type': '1',
                '@available': '1',
                'artist': 'artist',
                'album': 'album',
                'mediaid': 'mediaid',
                'tracklength': '0',
                'title': 'title',
                'contentid': '2',
                'thumbnail': 'http://thumb.url/image',
            }
        ]

        api.get_current_play_time.return_value = {
            'timelength': '192',
            'playtime': '35',
        }

        track = player.get_current_track()

        self.assertEqual(track.title, 'Don\'t Wait Too Long')
        self.assertEqual(track.artist, 'Madeleine Peyroux')
        self.assertEqual(track.album, 'Careless Love')
        self.assertEqual(track.duration, 192)
        self.assertEqual(track.position, 35)
        self.assertEqual(track.thumbnail_url, 'http://api.deezer.com/album/100127/image')
        self.assertEqual(track.object_id, '881851')
        self.assertEqual(track.object_type, 'app_audio')

    def test_is_active(self):
        player, api = _get_player()

        self.assertTrue(player.is_active('wifi', 'cp'))
        self.assertFalse(player.is_active('wifi', 'dlna'))
        self.assertFalse(player.is_active('bt'))



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.api import ApiResponse
from samsung_multiroom.api import ApiStream


def _get_api_stream():
    return ApiStream('public', '192.168.1.129')


class TestApiStream(unittest.TestCase):

    @unittest.mock.patch('socket.socket')
    def test_open_working_stream(self, s):
        s.return_value.recv.side_effect = [
            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 215
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>RequestDeviceInfo</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"></response></UIC>""",
            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 678
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>MainInfo</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier></user_identifier><response result="ok"><party>off</party><partymain></partymain><grouptype>N</grouptype><groupmainip>0.0.0.0</groupmainip><groupmainmacaddr>00:00:00:00:00:00</groupmainmacaddr><spkmacaddr>xx:xx:xx:xx:xx:xx</spkmacaddr><spkmodelname>HW-K650</spkmodelname><groupmode>none</groupmode><channeltype>front</channeltype><channelvolume>0</channelvolume><multichinfo>on</multichinfo><groupspknum>1</groupspknum><dfsstatus>dfsoff</dfsstatus><protocolver>2.3</protocolver><btmacaddr>yy:yy:yy:yy:yy:yy</btmacaddr></response></UIC>""",
        ]

        expected_responses = [
            {
                'success': True,
                'name': 'RequestDeviceInfo',
            },
            {
                'success': True,
                'name': 'MainInfo',
            }
        ]

        stream = _get_api_stream()

        responses = list(stream.open('/UIC?cmd=%3Cname%3EGetMainInfo%3C/name%3E'))

        s.return_value.connect.assert_called_once_with(('192.168.1.129', 55001))

        self.assertEqual(len(responses), len(expected_responses))

        for i, response in enumerate(responses):
            self.assertIsInstance(response, ApiResponse)
            self.assertEqual(response.success, expected_responses[i]['success'])
            self.assertEqual(response.name, expected_responses[i]['name'])


    @unittest.mock.patch('socket.socket')
    def test_open_multiple_responses_in_recv(self, s):
        s.return_value.recv.side_effect = [
            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 215
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>RequestDeviceInfo</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"></response></UIC>""",

            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 678
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>MainInfo</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier></user_identifier><response result="ok"><party>off</party><partymain></partymain><grouptype>N</grouptype><groupmainip>0.0.0.0</groupmainip><groupmainmacaddr>00:00:00:00:00:00</groupmainmacaddr><spkmacaddr>xx:xx:xx:xx:xx:xx</spkmacaddr><spkmodelname>HW-K650</spkmodelname><groupmode>none</groupmode><channeltype>front</channeltype><channelvolume>0</channelvolume><multichinfo>on</multichinfo><groupspknum>1</groupspknum><dfsstatus>dfsoff</dfsstatus><protocolver>2.3</protocolver><btmacaddr>yy:yy:yy:yy:yy:yy</btmacaddr></response></UIC>""",

            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 228
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>VolumeLevel</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"><volume>10</volume></response></UIC>HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 228
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>VolumeLevel</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"><volume>15</volume></response></UIC>HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 228
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>VolumeLevel</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"><volume>20</volume></response></UIC>""",

            b"""HTTP/1.1 200 OK
Date: Fri, 02 Jan 1970 10:53:13 GMT
Server: Samsung/1.0
Content-Type: text/html
Content-Length: 228
Connection: close
Last-Modified: Fri, 02 Jan 1970 10:53:13 GMT

<?xml version="1.0" encoding="UTF-8"?><UIC><method>VolumeLevel</method><version>1.0</version><speakerip>192.168.1.129</speakerip><user_identifier>public</user_identifier><response result="ok"><volume>30</volume></response></UIC>""",
        ]

        expected_responses = [
            {
                'success': True,
                'name': 'RequestDeviceInfo',
            },
            {
                'success': True,
                'name': 'MainInfo',
            },
            {
                'success': True,
                'name': 'VolumeLevel',
            },
            {
                'success': True,
                'name': 'VolumeLevel',
            },
            {
                'success': True,
                'name': 'VolumeLevel',
            },
            {
                'success': True,
                'name': 'VolumeLevel',
            }
        ]

        stream = _get_api_stream()

        responses = list(stream.open('/UIC?cmd=%3Cname%3EGetMainInfo%3C/name%3E'))

        s.return_value.connect.assert_called_once_with(('192.168.1.129', 55001))

        self.assertEqual(len(responses), len(expected_responses))

        for i, response in enumerate(responses):
            self.assertIsInstance(response, ApiResponse)
            self.assertEqual(response.success, expected_responses[i]['success'])
            self.assertEqual(response.name, expected_responses[i]['name'])



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service.dlna import DlnaBrowser
from samsung_multiroom.service.dlna import DlnaService


def _get_service():
    api = MagicMock()

    service = DlnaService(api)

    return (service, api)


class TestDlnaService(unittest.TestCase):

    def test_name(self):
        service, api = _get_service()

        self.assertEqual(service.name, 'dlna')

    def test_browser(self):
        service, api = _get_service()

        browser = service.browser

        self.assertIsInstance(browser, DlnaBrowser)

    def test_login(self):
        service, api = _get_service()

        service.login('test_username', 'test_password')

        api.set_sign_in.assert_not_called()

    def test_logout(self):
        service, api = _get_service()

        service.logout()

        api.set_sign_out.assert_not_called()



import unittest
from unittest.mock import MagicMock

import pytest

from samsung_multiroom.api import ApiResponse
from samsung_multiroom.event import EventLoop
from samsung_multiroom.event.event import Event


def _get_event_loop():
    api_stream = MagicMock()

    event_loop = EventLoop(api_stream)

    return (event_loop, api_stream)


def _fake_event_factory(response):
    event = MagicMock(spec=Event)
    event.name = 'fake.event'
    return event


def _get_api_response(name):
    event = MagicMock(spec=ApiResponse)
    event.name = name
    return event


class TestEventLoop():  # pytest-asyncio doesn't play well with unittest.TestCase

    @pytest.mark.asyncio
    async def test_loop(self):
        listener = MagicMock()

        event_loop, api_stream = _get_event_loop()

        api_stream.open.return_value = iter([
            _get_api_response('FakeEvent'),
        ])

        event_loop.register_factory(_fake_event_factory)
        event_loop.add_listener('fake.event', listener)

        await event_loop.loop()

        listener.assert_called_once()

    @pytest.mark.asyncio
    async def test_loop_no_events(self):
        listener = MagicMock()

        event_loop, api_stream = _get_event_loop()

        api_stream.open.return_value = iter([
            _get_api_response('FakeEvent'),
        ])

        event_loop.add_listener('*', listener)

        await event_loop.loop()

        listener.assert_not_called()



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.type.speaker_mute_changed import SpeakerMuteChangedEvent


class TestSpeakerMuteChangedEvent(unittest.TestCase):

    def test_factory(self):
        response = MagicMock()
        response.name = 'MuteStatus'
        response.data = {'mute': 'on'}

        event = SpeakerMuteChangedEvent.factory(response)

        self.assertIsInstance(event, SpeakerMuteChangedEvent)
        self.assertEqual(event.muted, True)

    def test_factory_returns_none(self):
        response = MagicMock()
        response.name = 'OtherResponse'

        event = SpeakerMuteChangedEvent.factory(response)

        self.assertEqual(event, None)



"""Speaker interface to control speaker operation."""
import abc


class SpeakerBase(metaclass=abc.ABCMeta):
    """Speaker interface to control speaker operation."""

    @property
    def ip_address(self):
        """
        :returns: Speaker's ip address
        """
        raise NotImplementedError()

    @property
    def mac_address(self):
        """
        :returns: Speaker's mac address
        """
        raise NotImplementedError()

    def get_name(self):
        """
        Retrieve speaker's name.

        :returns: Speaker name string
        """
        raise NotImplementedError()

    def set_name(self, name):
        """
        Set speaker's name.

        :param name: Speaker name string
        """
        raise NotImplementedError()

    def get_volume(self):
        """
        Get current speaker volume.

        :returns: int current volume
        """
        raise NotImplementedError()

    def set_volume(self, volume):
        """
        Set speaker volume.

        :param volume: speaker volume, integer between 0 and 100
        """
        raise NotImplementedError()

    def get_sources(self):
        """
        Get all supported sources.

        :returns: List of supported sources.
        """
        raise NotImplementedError()

    def get_source(self):
        """
        Get currently selected source.

        :returns: selected source string.
        """
        raise NotImplementedError()

    def set_source(self, source):
        """
        Set speaker source.

        :param source: Speaker source, one of returned by get_sources()
        """
        raise NotImplementedError()

    def is_muted(self):
        """
        Check if speaker is muted.

        :returns: True if muted, False otherwise
        """
        raise NotImplementedError()

    def mute(self):
        """Mute the speaker."""
        raise NotImplementedError()

    def unmute(self):
        """Unmute the speaker."""
        raise NotImplementedError()

    @property
    def event_loop(self):
        """
        Get event loop

        :returns: EventLoop instance
        """
        raise NotImplementedError()

    def get_services_names(self):
        """
        Get all supported services names.

        :returns: List of strings
        """
        raise NotImplementedError()

    def service(self, name):
        """
        Get service by type

        :returns: Service instance
        """
        raise NotImplementedError()

    @property
    def clock(self):
        """
        Get clock to control time functions.

        :returns: Clock instance
        """
        raise NotImplementedError()

    @property
    def equalizer(self):
        """
        Get equalizer to control sound adjustments.

        :returns: Equalizer instance
        """
        raise NotImplementedError()

    @property
    def player(self):
        """
        Get currently active player.

        Use player to control playback e.g. pause, resume, get track info etc.

        :returns: Player instance
        """
        raise NotImplementedError()

    def browser(self, name):
        """
        Get media browser by type

        :returns: Browser instance
        """
        raise NotImplementedError()

    def group(self, name, speakers):
        """
        Group this speaker with another ones.

        This speaker will be the main speaker controlling the playback.

        :param speaker: List of Speaker instances
        :returns: SpeakerGroup instance
        """
        raise NotImplementedError()

    def ungroup(self):
        """
        Remove this speaker from its current group.
        """
        raise NotImplementedError()



"""Event."""
from ..event import Event


class SpeakerServiceEvent(Event):
    """Event when service is changed."""

    def __init__(self, name, service_name):
        """
        :param name: Event name
        :param service_name: Name of the service selected
        """
        super().__init__(name)

        self._service_name = service_name

    @property
    def service_name(self):
        """
        :returns: Name of the service selected
        """
        return self._service_name

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerServiceEvent instance or None if response is unsupported
        """
        valid_responses_map = {
            'CpChanged': 'speaker.service.changed',
            'SignInStatus': 'speaker.service.logged_in',
            'SignOutStatus': 'speaker.service.logged_out',
        }

        if response.name not in valid_responses_map:
            return None

        return cls(valid_responses_map[response.name], response.data['cpname'])



"""Group multiple equalizers together to act on all simultaneously."""
from .equalizer import EqualizerBase


class EqualizerGroup(EqualizerBase):
    """
    Group multiple equalizers together to act on all simultaneously.
    """

    def __init__(self, equalizers):
        """
        :param equalizers: List of EqualizerBase implementations
        """
        self._equalizers = equalizers

    @property
    def equalizers(self):
        """
        :returns: List of equalizers
        """
        return self._equalizers

    def get_presets_names(self):
        """
        :returns: List of preset names shared by all equalizers.
        """
        presets_names = self._equalizers[0].get_presets_names()

        for equalizer in self._equalizers[1:]:
            eq_presets_names = equalizer.get_presets_names()
            presets_names = [p for p in presets_names if p in eq_presets_names]

        return presets_names

    def set(self, *args):
        """
        Set equalizer values by preset name or list of band values for all equalizers.

        :param: Name of the preset to set or list of 7 band values to set
        """
        for equalizer in self._equalizers:
            equalizer.set(*args)

    def save(self, name=None):
        """
        Create a new or overwrite existing preset for all equalizers.

        :param name: If empty, current preset will be saved, if provided name exist, that preset will overwrite existing
            one, otherwise a new preset is created with that name. In either case band values used will be ones
            currently set on a speaker e.g. via set() method.
        """
        for equalizer in self._equalizers:
            equalizer.save(name)

    def delete(self, name):
        """
        Delete preset from all equalizers.

        :param name: Name of a preset to delete
        """
        for equalizer in self._equalizers:
            equalizer.delete(name)



"""
Init.
"""
# pylint: disable=C0103
from . import factory
from . import discovery
from .service import REPEAT_ONE, REPEAT_ALL, REPEAT_OFF

# aliases
SamsungMultiroomSpeaker = factory.speaker_factory
SamsungSpeakerDiscovery = discovery.SpeakerDiscovery



"""Speaker music service integrations."""
import abc


class Service(metaclass=abc.ABCMeta):
    """
    Abstract service
    """

    def __init__(self, browser):
        """
        :param browser: Browser instance specific to implementing service
        """
        self._browser = browser

    @property
    @abc.abstractmethod
    def name(self):
        """
        :returns: Name of the service
        """
        raise NotImplementedError()

    @property
    def browser(self):
        """
        :returns: Compatible Browser instance for this service
        """
        return self._browser

    def login(self, username, password):
        """
        Authenticate with username/password.

        :param username:
        :param password:
        """

    def logout(self):
        """
        Reset authentication state.
        """



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.event import Event


class TestEvent(unittest.TestCase):

    def test_name(self):
        event = Event('speaker.volume.changed')

        self.assertEqual(event.name, 'speaker.volume.changed')



import datetime
import unittest
from unittest.mock import MagicMock

from samsung_multiroom.clock import Clock


def get_clock():
    api = MagicMock()
    alarm = MagicMock()
    timer = MagicMock()

    clock = Clock(api, timer, alarm)

    return (clock, api, alarm, timer)


NOW = datetime.datetime(2018, 1, 7, 15, 45, 32)


class TestClock(unittest.TestCase):

    def test_set_time(self):
        clock, api, alarm, timer = get_clock()

        spk_datetime = datetime.datetime.now()

        clock.set_time(spk_datetime)

        api.set_speaker_time.assert_called_once_with(spk_datetime)

    @unittest.mock.patch('datetime.datetime')
    def test_set_time_now(self, dt):
        clock, api, alarm, timer = get_clock()

        dt.now.return_value = NOW

        clock.set_time()

        api.set_speaker_time.assert_called_once_with(NOW)

    def test_timer(self):
        clock, api, alarm, timer = get_clock()

        self.assertEqual(clock.timer, timer)

    def test_alarm(self):
        clock, api, alarm, timer = get_clock()

        self.assertEqual(clock.alarm, alarm)



import datetime
import unittest
from unittest.mock import MagicMock

from samsung_multiroom.clock import ClockGroup


def _get_clock_group():
    clocks = [
        MagicMock(),
        MagicMock(),
        MagicMock(),
    ]

    clock_group = ClockGroup(clocks)

    return (clock_group, clocks)


class TestClockGroup(unittest.TestCase):

    def test_set_time_sets_all_clocks(self):
        clock_group, clocks = _get_clock_group()

        dt = datetime.datetime.now()

        clock_group.set_time(dt)

        for clock in clocks:
            clock.set_time.assert_called_once_with(dt)

    def test_alarm_returns_first_clock_alarm(self):
        clock_group, clocks = _get_clock_group()
        clocks[0].alarm = MagicMock()

        alarm = clock_group.alarm

        self.assertEqual(alarm, clocks[0].alarm)

    def test_timer_returns_first_clock_timer(self):
        clock_group, clocks = _get_clock_group()
        clocks[0].timer = MagicMock()

        timer = clock_group.timer

        self.assertEqual(timer, clocks[0].timer)



"""Control alarm functions of the speaker to wake speaker at specific time."""
import datetime


class Alarm:
    """
    Control alarm functions of the speaker to wake speaker at specific time.
    """

    def __init__(self, api):
        """
        :param api: SamsungMultiroomApi instance
        """
        self._api = api
        self._slots = [None, None, None]

    def __getitem__(self, i):
        return self.get_slots()[i]

    def slot(self, i):
        """
        Get slot of a specific index.

        :param i: Slot index, must be between 0 and 2
        :returns: AlarmSlot instance
        """
        if i < 0 or i > 2:
            raise ValueError('Only three slots are available')

        return self.get_slots()[i]

    def get_slots(self):
        """
        Get alarm slots.

        Speakers have 3 alarm slots available. This method will return the ones that are set as well as empty ones to
        use for setting new alarms.

        :returns: List of AlarmSlot instances
        """
        alarms = self._api.get_alarm_info()

        for alarm in alarms:
            index = int(alarm['@index'])
            self._slots[index] = AlarmSlot(self._api, index, alarm)

        # fill with empty slots
        for index in range(3):
            if self._slots[index] is None:
                self._slots[index] = AlarmSlot(self._api, index)

        return self._slots


class AlarmSlot:
    """
    Specific alarm configuration.

    Speakers are limited to have only three alarms set at any given time. Alarm slot represents a single alarm,
    configured or not.
    """

    def __init__(self, api, index, alarm_info=None):
        """
        :param api: SamsungMultiroomApi instance
        :param index: Slot index
        """
        self._api = api
        self._index = index

        if alarm_info:
            self._time = '{0}:{1:02d}'.format(int(alarm_info['hour']), int(alarm_info['min']))
            self._weekdays = hexweek_to_weekday_list(alarm_info['week'])
            self._volume = int(alarm_info['volume'])
            self._duration = int(alarm_info['duration'])
            self._enabled = alarm_info['set'] == 'on'
            self._station_data = {
                'title': alarm_info['title'],
                'description': alarm_info['description'],
                'thumbnail_url': alarm_info['thumbnail'],
                'station_url': alarm_info['stationurl'],
            }
        else:
            self._set_defaults()

    def _set_defaults(self):
        self._time = datetime.datetime.now().strftime('%H:%M')
        self._weekdays = [datetime.datetime.now().weekday()]
        self._duration = 0
        self._volume = 5
        self._station_data = None
        self._enabled = False

    def set(self, time=None, weekdays=None, duration=None, volume=None, station_data=None, playlist=None, enabled=None):
        """
        Update speaker's configuration for this slot.

        :param time: hour/minute when to trigger the alarm in HH:MM format
        :param weekdays: List of integers representing weekdays when alarm is triggered. 0 - Mon, 1 - Tue, ... 6 - Sun
        :param duration: Duration of alarm in seconds
        :param volume: Volume to set when alarm is triggered between 0 and 100
        :param station_data: Alarm sound, must be dict:
            - title
            - description
            - thumbnail_url
            - station_url
        :param playlist: Alarm sound, must be iterable returning alarm combatible objects
            - object_type - must be set to tunein_radio
            - object_id - object id
            - title - radio name
        :param enabled: True if alarm should be active, False otherwise
        """
        if time:
            self._time = time
        if weekdays:
            self._weekdays = weekdays
        if duration:
            self._duration = int(duration)
        if volume:
            self._volume = int(volume)
        if station_data:
            self._station_data = station_data
        elif playlist:
            self._station_data = self._get_station_data_from_playlist(playlist)
        elif not self._station_data:
            raise ValueError('Either station_data or playlist must be provided')
        if enabled is not None:
            self._enabled = bool(enabled)

        self._api.set_alarm_info(
            index=self._index,
            hour=int(self._time.split(':')[0]),
            minute=int(self._time.split(':')[1]),
            week=weekday_list_to_hexweek(self._weekdays),
            station_data={
                'title': self._station_data['title'],
                'description': self._station_data['description'],
                'thumbnail': self._station_data['thumbnail_url'],
                'stationurl': self._station_data['station_url'],
            },
            volume=self._volume,
            duration=self._duration)

        self._api.set_alarm_on_off(self._index, 'on' if self._enabled else 'off')

    def delete(self):
        """
        Delete this alarm and set alarm settings to defaults.
        """
        self._api.del_alarm(self._index)
        self._set_defaults()

    def disable(self):
        """
        Disable this alarm.
        """
        self._api.set_alarm_on_off(self._index, 'off')
        self._enabled = False

    def enable(self):
        """
        Enable this alarm.
        """
        self._api.set_alarm_on_off(self._index, 'on')
        self._enabled = True

    def _get_station_data_from_playlist(self, playlist):
        """
        Get station_data by providing playlist of compatible items.

        Playlist items must be an object with following attributes:
        - object_id - object id
        - object_type - must be 'tunein_radio'
        - title - radio name

        :param playlist: Iterable returning alarm combatible objects
        """
        for radio in playlist:
            if radio.object_type not in ['tunein_radio']:
                continue

            station_data = self._api.get_station_data(radio.object_id)
            return {
                'title': station_data['title'] or '',
                'description': station_data['description'] or '',
                'thumbnail_url': station_data['thumbnail'] or '',
                'station_url': station_data['stationurl'] or '',
            }

        raise ValueError('No compatible playlist items. Object type must be tunein_radio.')

    @property
    def index(self):
        """
        :returns: Slot index between 0 and 2
        """
        return self._index

    @property
    def time(self):
        """
        :returns: Alarm trigger time string in HH:MM format
        """
        return self._time

    @property
    def weekdays(self):
        """
        :returns: List of integers representing weekdays when alarm is triggered. 0 - Mon, 1 - Tue, ... 6 - Sun
        """
        return self._weekdays

    @property
    def duration(self):
        """
        :returns: Alarm duration in seconds
        """
        return self._duration

    @property
    def volume(self):
        """
        :returns: Speaker volume when alarm is triggered
        """
        return self._volume

    @property
    def station_data(self):
        """
        :returns: radio station data dict that is used as alarm sound
            - name
            - description
            - thumbnail_url
            - station_url
        """
        return self._station_data

    @property
    def enabled(self):
        """
        :returns: True if alarm is enabled
        """
        return self._enabled


def hexweek_to_weekday_list(hexweek):
    """
    Helper to convert speaker's hex representation of weekdays into list of integers represending weekdays.

    :param hexweek: Hex string .e.g. 0x3E
    :returns: List of weekday integers e.g. [0, 1, 2, 3, 4] for hexweek 0x3E
    """
    intweek = int(hexweek, 16)

    # Mon, Tue, Wed, Thu, Fri, Sat, Sun
    weekday_bits = [32, 16, 8, 4, 2, 1, 64]

    return [weekday for weekday, weekday_bit in enumerate(weekday_bits) if intweek & weekday_bit]


def weekday_list_to_hexweek(weekday_list):
    """
    Helper to convert list of integers represending weekdays into speaker's hex representation of weekdays.

    :param hexweek: List of weekday integers e.g. [0, 1, 2, 3, 4]
    :returns: Hex string .e.g. 0x3E
    """
    # Mon, Tue, Wed, Thu, Fri, Sat, Sun
    weekday_bits = [32, 16, 8, 4, 2, 1, 64]
    weekday_list = set(weekday_list)

    return hex(sum([weekday_bits[weekday] for weekday in weekday_list]))



"""DLNA service player."""
from ..player import REPEAT_ALL
from ..player import REPEAT_OFF
from ..player import REPEAT_ONE
from ..player import Player
from ..player import Track
from ..player import init_track_kwargs
from ..player import unsupported


class DlnaPlayer(Player):
    """Controls player in WIFI+DLNA mode."""

    def __init__(self, api):
        self._api = api

    def play(self, playlist):
        """
        Enqueue and play a playlist.

        Playlist items must be an object with following attributes:
        - object_id - object id
        - object_type - must be 'dlna_audio'
        - title - track title
        - artist - track artist
        - thumbnail_url - thumbnail URL
        - device_udn - DLNA device UDN

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        items = []
        for track in playlist:
            if track.object_type not in ['dlna_audio']:
                continue

            items.append({
                'object_id': track.object_id,
                'title': track.title,
                'artist': track.artist,
                'thumbnail': track.thumbnail_url,
                'device_udn': track.device_udn,
            })

        if items:
            self._api.set_playlist_playback_control(items)
            return True

        return False

    def jump(self, time):
        """
        Advance current playback to specific time.

        :param time: Time from the beginning of the track in seconds
        """
        self._api.set_search_time(time)

    def resume(self):
        """Play/resume current track."""
        self._api.set_playback_control('resume')

    @unsupported
    def stop(self):
        """Stop current track and reset position to the beginning."""
        raise NotImplementedError()

    def pause(self):
        """Pause current track and retain position."""
        self._api.set_playback_control('pause')

    def next(self):
        """Play next track in the queue."""
        self._api.set_trick_mode('next')

    def previous(self):
        """Play previous track in the queue."""
        self._api.set_trick_mode('previous')

    def repeat(self, mode):
        """
        Set playback repeat mode.

        :param mode: one of REPEAT_* constants
        """
        mode_map = {
            REPEAT_ALL: 'all',
            REPEAT_ONE: 'one',
            REPEAT_OFF: 'off',
        }

        if mode not in mode_map:
            raise ValueError('mode must be one of REPEAT_* constants')

        self._api.set_repeat_mode(mode_map[mode])

    def shuffle(self, enabled):
        """
        Enable/disable playback shuffle mode.

        :param enabled: True to enable, False to disable
        """
        self._api.set_shuffle_mode(enabled)

    def get_repeat(self):
        """
        Get playback repeat mode.

        :returns: one of REPEAT_* constants
        """
        mode_map = {
            'all': REPEAT_ALL,
            'one': REPEAT_ONE,
            'off': REPEAT_OFF,
        }

        mode = self._api.get_repeat_mode()

        return mode_map[mode]

    def get_shuffle(self):
        """
        Get playback shuffle mode.

        :returns: boolean, True if enabled, False otherwise
        """
        return self._api.get_shuffle_mode()

    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        music_info = self._api.get_music_info()

        track_kwargs = init_track_kwargs('dlna_audio')

        if 'title' in music_info:
            track_kwargs['title'] = music_info['title']
        if 'artist' in music_info:
            track_kwargs['artist'] = music_info['artist']
        if 'album' in music_info:
            track_kwargs['album'] = music_info['album']
        if 'thumbnail' in music_info and 'http' in music_info['thumbnail']:
            track_kwargs['thumbnail_url'] = music_info['thumbnail']
        if 'timelength' in music_info and music_info['timelength'] is not None:
            (hours, minutes, seconds) = music_info['timelength'].split(':')
            track_kwargs['duration'] = int(hours) * 3600 + int(minutes) * 60 + int(float(seconds))
        if 'playtime' in music_info:
            track_kwargs['position'] = int(int(music_info['playtime']) / 1000)
        if 'device_udn' in music_info:
            track_kwargs['metadata']['device_udn'] = music_info['device_udn']
        if 'objectid' in music_info:
            track_kwargs['metadata']['object_id'] = music_info['objectid']

        return Track(**track_kwargs)

    def is_active(self, function, submode=None):
        """
        Check if this player is active based on current function/submode.

        :returns: Boolean True if function/submode is supported
        """
        return function == 'wifi' and submode == 'dlna'



"""Event."""
from ..event import Event


class SpeakerPlayerEvent(Event):
    """Event when player status changes."""

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerPlayerEvent instance or None if response is unsupported
        """
        valid_responses_map = {
            'StartPlaybackEvent': 'speaker.player.playback_started',
            'StopPlaybackEvent': 'speaker.player.playback_ended',
            'EndPlaybackEvent': 'speaker.player.playback_ended',
            'PausePlaybackEvent': 'speaker.player.playback_paused',
            'MediaBufferStartEvent': 'speaker.player.buffering_started',
            'MediaBufferEndEvent': 'speaker.player.buffering_ended',
        }

        if response.name not in valid_responses_map:
            return None

        return cls(valid_responses_map[response.name])



import unittest
from unittest.mock import MagicMock

from samsung_multiroom import SamsungMultiroomSpeaker
from samsung_multiroom.group import SpeakerGroup
from samsung_multiroom.speaker import Speaker


def get_speaker():
    api = MagicMock()
    api.ip_address = '192.168.1.129'

    clock = MagicMock()
    equalizer = MagicMock()
    player_operator = MagicMock()

    service_registry = MagicMock()

    event_loop = MagicMock()

    speaker = Speaker(api, event_loop, clock, equalizer, player_operator, service_registry)

    return (speaker, api, event_loop, clock, equalizer, player_operator, service_registry)


class TestSpeaker(unittest.TestCase):

    def test_factory(self):
        self.assertIsInstance(SamsungMultiroomSpeaker('192.168.1.129'), Speaker)

    def test_get_name(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        api.get_speaker_name.return_value = 'Speaker name'

        name = speaker.get_name()

        api.get_speaker_name.assert_called_once()
        self.assertEqual(name, 'Speaker name')

    def test_set_name(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.set_name('Living Room')

        api.set_speaker_name.assert_called_once_with('Living Room')

    def test_get_volume(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        api.get_volume.return_value = 10

        volume = speaker.get_volume()

        api.get_volume.assert_called_once()
        self.assertEqual(volume, 10)

    def test_set_volume(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.set_volume(10)

        api.set_volume.assert_called_once_with(10)

    def test_get_sources(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        sources = speaker.get_sources()

        self.assertEqual(sorted(sources), sorted(['aux', 'bt', 'hdmi', 'optical', 'soundshare', 'wifi']))

    def test_get_source(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        api.get_func.return_value = {'function':'wifi', 'submode':'dlna'}

        source = speaker.get_source()

        api.get_func.assert_called_once()
        self.assertEqual(source, 'wifi')

    def test_set_source(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.set_source('hdmi')

        api.set_func.assert_called_once_with('hdmi')

    def test_is_muted(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        api.get_mute.return_value = True

        muted = speaker.is_muted()

        api.get_mute.assert_called_once()
        self.assertTrue(muted)

    def test_mute(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.mute()

        api.set_mute.assert_called_once_with(True)

    def test_unmute(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.unmute()

        api.set_mute.assert_called_once_with(False)

    def test_event_loop(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        el = speaker.event_loop

        self.assertEqual(el, event_loop)

    def test_get_services_names(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        service_registry.get_services_names.return_value = ['dlna', 'tunein', 'deezer', 'spotify']

        services_names = speaker.get_services_names()

        self.assertEqual(services_names, ['dlna', 'tunein', 'deezer', 'spotify'])
        service_registry.get_services_names.assert_called_once()

    def test_service(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()
        service_registry.service.return_value = MagicMock()

        service = speaker.service('dlna')

        self.assertEqual(service, service_registry.service.return_value)
        service_registry.service.assert_called_once_with('dlna')

    def test_player(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        player = speaker.player

        self.assertEqual(player, player_operator)

    def test_browser(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        browser = MagicMock()
        service_registry.service.return_value = MagicMock(browser=browser)

        service_browser = speaker.browser('b2')

        self.assertEqual(service_browser, browser)
        service_registry.service.assert_called_once_with('b2')

    def test_equalizer(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        eq = speaker.equalizer

        self.assertEqual(eq, equalizer)

    def test_clock(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        cl = speaker.clock

        self.assertEqual(cl, clock)

    def test_group(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker1 = MagicMock()
        speaker2 = MagicMock()

        api.get_speaker_name.return_value = 'This speaker'
        api.get_main_info.return_value = {
            'spkmacaddr': '00:00:00:00:00:00'
        }

        speaker1.get_name.return_value = 'Speaker 1'
        speaker2.get_name.return_value = 'Speaker 2'

        speaker1.ip_address = '192.168.1.165'
        speaker2.ip_address = '192.168.1.216'

        speaker1.mac_address = '11:11:11:11:11:11'
        speaker2.mac_address = '22:22:22:22:22:22'

        group = speaker.group('My group', [speaker1, speaker2])

        self.assertIsInstance(group, SpeakerGroup)
        api.set_multispk_group.assert_called_once_with('My group', [
            {
                'name': 'This speaker',
                'ip': '192.168.1.129',
                'mac': '00:00:00:00:00:00',
            },
            {
                'name': 'Speaker 1',
                'ip': '192.168.1.165',
                'mac': '11:11:11:11:11:11',
            },
            {
                'name': 'Speaker 2',
                'ip': '192.168.1.216',
                'mac': '22:22:22:22:22:22',
            }
        ])

    def test_ungroup(self):
        speaker, api, event_loop, clock, equalizer, player_operator, service_registry = get_speaker()

        speaker.ungroup()

        api.set_ungroup.assert_called_once()



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.type.speaker_service import SpeakerServiceEvent


class TestSpeakerPlayerEvent(unittest.TestCase):

    def test_factory(self):
        responses = {
            'CpChanged': 'speaker.service.changed',
            'SignInStatus': 'speaker.service.logged_in',
            'SignOutStatus': 'speaker.service.logged_out',
            'OtherResponse': None,
        }

        for response_name, event_name in responses.items():
            response = MagicMock()
            response.name = response_name
            response.data = {'cpname': 'MyService'}

            event = SpeakerServiceEvent.factory(response)

            if event_name:
                self.assertIsInstance(event, SpeakerServiceEvent)
                self.assertEqual(event.name, event_name)
                self.assertEqual(event.service_name, 'MyService')
            else:
                self.assertIsNone(event)



"""TuneIn service."""
from ..service import Service
from .browser import TuneInBrowser


class TuneInService(Service):
    """
    TuneIn radio service.
    """

    def __init__(self, api):
        super().__init__(TuneInBrowser(api))

    @property
    def name(self):
        """
        :returns: Name of the service
        """
        return 'tunein'



"""Discover speakers on your local network."""
from .factory import speaker_factory


class SpeakerDiscovery:
    """
    Discover speakers on your local network.
    """

    def __init__(self):
        """Init."""
        self._speakers = {}

    def discover(self):
        """
        Discover speakers.

        :returns: List of Speaker instances
        """
        import upnpclient
        from urllib.parse import urlparse

        devices = upnpclient.discover()
        for device in devices:
            if not self._is_compatible_device(device):
                continue

            url = urlparse(device.location)
            hostname = url.hostname

            if hostname in self._speakers:
                continue

            self._speakers[hostname] = speaker_factory(hostname)

        return list(self._speakers.values())

    def _is_compatible_device(self, device):
        """
        Samsung speakers report on /smp_3_ and /smp_7_
        Samsung TVs report on /smp_2_ /smp_7_ /smp_15_ /smp_25_

        Therefore /smp_3_ seems to be specific to Samsung speakers.
        """
        valid_services = [s for s in device.services if s.service_id == 'urn:samsung.com:serviceId:MultiScreenService']
        valid_location = '/smp_3_' in device.location

        return valid_services and valid_location



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.type.speaker_volume_changed import SpeakerVolumeChangedEvent


class TestSpeakerVolumeChangedEvent(unittest.TestCase):

    def test_factory(self):
        response = MagicMock()
        response.name = 'VolumeLevel'
        response.data = {'volume': '25'}

        event = SpeakerVolumeChangedEvent.factory(response)

        self.assertIsInstance(event, SpeakerVolumeChangedEvent)
        self.assertEqual(event.volume, 25)

    def test_factory_returns_none(self):
        response = MagicMock()
        response.name = 'OtherResponse'

        event = SpeakerVolumeChangedEvent.factory(response)

        self.assertEqual(event, None)



"""Event."""
from ..event import Event


class SpeakerVolumeChangedEvent(Event):
    """Event when speaker volume is adjusted."""

    def __init__(self, volume):
        """
        :param volume: int new volume
        """
        super().__init__('speaker.volume.changed')

        self._volume = volume

    @property
    def volume(self):
        """
        :returns: new volume
        """
        return self._volume

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerVolumeChangedEvent instance or None if response is unsupported
        """
        if response.name != 'VolumeLevel':
            return None

        return cls(int(response.data['volume']))



"""
Stream messages from the speaker.
"""
import logging
import socket

from .api_response import ApiResponse

try:
    from http_parser.parser import HttpParser
except ImportError:
    from http_parser.pyparser import HttpParser

_LOGGER = logging.getLogger(__name__)


class ApiStream:
    """
    Speaker's api stream.

    It is possible to listen to all events/responses a speaker generates. It is useful in situations where there are
    multiple clients operating the speaker. In such case you can maintain internal state without polling.

    Once opened it will listen to messages in definitely, until interrupted using close() method.

    Example:
        stream = ApiStream('unique-id', '129.168.1.129')

        for response in stream.open('/UIC?cmd=%3Cname%3EGetMainInfo%3C/name%3E'):
            print(response.data)
    """

    def __init__(self, user, ip_address, port=55001, timeout=None):
        """
        Initialise stream.

        :param user: User identifier to pass along with request
        :param ip_address: IP address of the speaker to connect to
        :param port: Port to use, defaults to 55001
        :param timeout: Timeout in seconds
        """
        self._user = user
        self._ip_address = ip_address
        self._port = port
        self._timeout = timeout
        self._continue_stream = False

    def open(self, uri):
        """
        Generator consuming events from speaker's main info stream.

        Yields ApiResponse instance.

        :param uri: URI to open for the stream e.g. /UIC?cmd=%3Cname%3EGetMainInfo%3C/name%3E
        """
        self._continue_stream = True

        headers = {
            'mobileUUID': self._user,
            'mobileName': 'Wireless Audio',
            'mobileVersion': '1.0',
        }

        while self._continue_stream:
            _LOGGER.debug('Opening new stream')
            try:
                parser = HttpParser()
                body = []

                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((self._ip_address, self._port))
                sock.settimeout(self._timeout)
                sock.send('GET {0} HTTP/1.1\r\n'.format(uri).encode())
                sock.send('Host: {0}:{1}\r\n'.format(self._ip_address, self._port).encode())
                for header, value in headers.items():
                    sock.send('{0}: {1}\r\n'.format(header, value).encode())
                sock.send('\r\n\r\n'.encode())

                while self._continue_stream:
                    _LOGGER.debug('Receiving from stream')
                    data = sock.recv(1024)

                    while data:
                        _LOGGER.debug('Received data: %s', data.decode())
                        received_length = len(data)
                        parsed_length = parser.execute(data, received_length)

                        if parser.is_partial_body():
                            body.append(parser.recv_body().decode())

                        if parser.is_message_complete():
                            full_body = ''.join(body)
                            parser = HttpParser()
                            body = []

                            _LOGGER.debug('Stream response: %s', full_body)

                            yield ApiResponse(full_body)

                        data = data[parsed_length:]
            except socket.error:
                _LOGGER.error('Socket exception', exc_info=1)
            finally:
                _LOGGER.debug('Closing the stream')
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()

    def close(self):
        """
        Attempt to interrupt currently open stream.
        """
        _LOGGER.debug('Requested to close the stream')
        self._continue_stream = False



"""Event."""
from ..event import Event


class SpeakerPlayerShuffleChangedEvent(Event):
    """Event when player's shuffle state changes."""

    def __init__(self, shuffle):
        """
        :param shuffle: Boolean
        """
        super().__init__('speaker.player.shuffle.changed')

        self._shuffle = shuffle

    @property
    def shuffle(self):
        """
        :returns: Shuffle state, True if shuffle is on
        """
        return self._shuffle

    @classmethod
    def factory(cls, response):
        """
        Factory event from response.

        :returns: SpeakerPlayerShuffleChangedEvent instance or None if response is unsupported
        """
        if response.name != 'ShuffleMode':
            return None

        return cls(response.data['shuffle'] == 'on')



"""Control time functions for the group of the speakers."""
from .clock import ClockBase


class ClockGroup(ClockBase):
    """
    Control time functions for the group of the speakers.

    Due to the nature of alarms and timers, it is only required to use those on the main speaker of the group.
    """

    def __init__(self, clocks):
        """
        :param clocks: List of Clock instances
        """
        self._clocks = clocks

    @property
    def clocks(self):
        """
        :returns: List of Clock instances in group
        """
        return self._clocks

    def set_time(self, speaker_datetime=None):
        """
        Set current time for all speakers in group.

        :param speaker_datetime: Datetime object
        """
        for clock in self._clocks:
            clock.set_time(speaker_datetime)

    @property
    def alarm(self):
        """
        It is not possible to update alarm functions while the speaker is part of a group.

        :returns: Alarm instance of the first clock
        """
        return self._clocks[0].alarm

    @property
    def timer(self):
        """
        :returns: Timer instance of the first clock
        """
        return self._clocks[0].timer



"""
Init.
"""
from .api import SamsungMultiroomApi
from .api import SamsungMultiroomApiException
from .api import paginator
from .api_stream import ApiStream
from .api_response import ApiResponse

from .api import COMMAND_CPM, COMMAND_UIC
from .api import METHOD_GET



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.event.type.speaker_player_repeat_changed import SpeakerPlayerRepeatChangedEvent
from samsung_multiroom.service.player import REPEAT_ALL


class TestSpeakerrepeatChangedEvent(unittest.TestCase):

    def test_factory(self):
        response = MagicMock()
        response.name = 'RepeatMode'
        response.data = {'repeat': 'all'}

        event = SpeakerPlayerRepeatChangedEvent.factory(response)

        self.assertIsInstance(event, SpeakerPlayerRepeatChangedEvent)
        self.assertEqual(event.repeat, REPEAT_ALL)

    def test_factory_returns_none(self):
        response = MagicMock()
        response.name = 'OtherResponse'

        event = SpeakerPlayerRepeatChangedEvent.factory(response)

        self.assertEqual(event, None)
